from utilities import load_save, metrics  # method_comparison
import os
from timeit import default_timer as timer
from main import BurjKhalifaClustering

SKIP_COLUMNS_IN_DATASET = {
    ("real", "wdbc"): ["idnumber"],
    ("real", "spectrometer"): ["LRS-name"],
}


def run_multiple_datasets(
    dataset_names,
    depth=2,
    chain_ratio=5,
    parent_split_ratio=10,
    min_leaves=0,
    n_clusters=None,
):
    result_dict = {}
    pred_dict = {}
    for idx, dataset_name in enumerate(dataset_names):
        try:
            start = timer()
            skip_columns = (
                []
                if dataset_name not in SKIP_COLUMNS_IN_DATASET
                else SKIP_COLUMNS_IN_DATASET[dataset_name]
            )
            df = load_save.read_arff(dataset_name[0], dataset_name[1], skip_columns)
            X, true_labels = df.loc[:, df.columns != "class"], df["class"]
            bk_model = BurjKhalifaClustering(
                depth=depth,
                chain_ratio=chain_ratio,
                parent_split_ratio=parent_split_ratio,
                min_leaves=min_leaves,
                n_clusters=n_clusters,
                linkage="ward",
            )
            bk_model.fit(X)
            predict_labels = bk_model.labels_
            error_results = metrics.calculate_metrics(true_labels, predict_labels)
            result_dict[dataset_name] = load_save.format_results(
                error_results, timer() - start
            )
            pred_dict[dataset_name] = load_save.encode_json(predict_labels)
        except Exception as e:
            print(e)
    load_save.save_json(
        pred_dict, "./../results/predictions/bk_clustering_predictions.json"
    )
    return result_dict


def run_batch():
    skip_datasets = [
        "water-treatment",  # no class
        "autos",
        "credit.a",  # duplicate dataset
        "credit.g",  # duplicate dataset
        "sick",  # duplicate dataset
        "golfball",  # as 1 cluster, incorrect metric definition for clustering methods
        "Colon",  # multiple duplicated column names
        "jm1",  # gmm throws error
        "KDDTest+",  # gmm throws error
        "Rice_MSC_Dataset",  # run separately
        "click_data",  # takes forever long for kmeans?
    ]
    folders = ["real", "artificial"]

    dataset_names = []
    for folder in folders:
        dataset_names += [
            (folder, x[:-5])
            for x in os.listdir(f"./../data/{folder}")
            if x[:-5] not in skip_datasets
        ]

    number_of_clusters = [
        load_save.read_arff(x[0], x[1]).iloc[:, -1].nunique() for x in dataset_names
    ]

    # run bk_clustering

    results = run_multiple_datasets(dataset_names)
    filename = f"./../results/bk_clustering_results.json"  # Define the filename to save the results
    load_save.save_json(results, filename)  # Save the results to a JSON file

    # method_comparison.run_birch(dataset_names, number_of_clusters)
    # method_comparison.run_dbscan(dataset_names, number_of_clusters)
    # method_comparison.run_kmeans(dataset_names, number_of_clusters)
    # method_comparison.run_kmeans_mini_batch(dataset_names, number_of_clusters)
    # method_comparison.run_mean_shift(dataset_names, number_of_clusters)
    # method_comparison.run_agglomerative(dataset_names, number_of_clusters)
    # method_comparison.run_gmm(dataset_names, number_of_clusters)
    # method_comparison.run_affinity_propagation(dataset_names, number_of_clusters)
    # method_comparison.run_density_peak(dataset_names, number_of_clusters)
    # method_comparison.run_optics(dataset_names, number_of_clusters)
    # method_comparison.run_hdbscan(dataset_names, number_of_clusters)


# python3 run_batch.py
if __name__ == "__main__":
    run_batch()



from typing import List, Tuple
import numpy as np
from sklearn.metrics import (
    rand_score,
    adjusted_rand_score,
    mutual_info_score,
    normalized_mutual_info_score,
    adjusted_mutual_info_score,
    homogeneity_score,
    completeness_score,
    v_measure_score,
    fowlkes_mallows_score,
)


def calculate_metrics(true_labels, pred_labels):
    """
    Calculate various clustering evaluation metrics based on true and predicted labels.

    Args:
        true_labels (List): List of true labels.
        pred_labels (List): List of predicted labels.

    Returns:
        dict: A dictionary containing various clustering evaluation metrics.
    """
    # Get the mapping of predicted labels to true labels and a boolean array indicating correct predictions
    map_dict, correct_list = get_correct_list(true_labels, pred_labels)

    # Calculate Rand Index
    rand_index = calculate_rand_index(true_labels, pred_labels)

    # Calculate Mutual Similarity
    mutual_similarity = calculate_mutual_similarity(true_labels, pred_labels)

    # Calculate V-Measure
    v_measure = calculate_v_measure(true_labels, pred_labels)

    # Create a dictionary to store the calculated metrics
    errors_dict = {
        "clusters_detected": len(map_dict),
        "clusters_true": len(set(true_labels)),
        "correctly_detected": int(sum(correct_list)),
        "number_of_datapoints": len(correct_list),
        "mutual_similarity_0": mutual_similarity[0],
        "mutual_similarity_1": mutual_similarity[1],
        "mutual_similarity_2": mutual_similarity[2],
        "rand_index_0": rand_index[0],
        "rand_index_1": rand_index[1],
        "v_measure_0": v_measure[0],
        "v_measure_1": v_measure[1],
        "v_measure_2": v_measure[2],
        "fm_score": calculate_fm_score(true_labels, pred_labels),
    }

    return errors_dict


def get_correct_list(true_labels: List, pred_labels: List):
    """
    Get the mapping of predicted labels to true labels and a boolean array indicating correct predictions.

    Args:
        true_labels (List): List of true labels.
        pred_labels (List): List of predicted labels.

    Returns:
        tuple: A tuple containing:
            - map_dict (dict): A dictionary mapping predicted labels to true labels.
            - correct (np.array): A boolean array indicating correct predictions.
    """
    # Initialize a nested dictionary to store counts of true labels for each predicted label
    res = {x: {y: 0 for y in set(true_labels)} for x in set(pred_labels)}

    # Update the counts in the nested dictionary based on the true and predicted labels
    for r, v in zip(pred_labels, true_labels):
        res[r][v] += 1

    # Get the mapping of predicted labels to true labels by selecting the label with the highest count
    map_dict = {r: max(res[r], key=res[r].get) for r in res}

    # Create a boolean array indicating correct predictions by comparing predicted labels with mapped true labels
    correct = np.array(true_labels) == np.array([map_dict[x] for x in pred_labels])

    return map_dict, correct


def calculate_v_measure(
    true_labels: List, pred_labels: List, beta: float = 1.0
) -> Tuple[float, float, float]:
    """
    The V-measure metric evaluates the accuracy of cluster assignments using conditional entropy analysis.
    A higher score indicates a higher degree of similarity.
    The V-measure metric comprises two intuitive sub-metrics that emerge from supervised learning:
        Homogeneity: This sub-metric evaluates whether each cluster comprises only members of a single class
            (similar to "precision").
        Completeness: This sub-metric assesses whether all members of a specific class are assigned to the
            same cluster (similar to "recall").

    Use:
    - When you require interpretability since V-measure is intuitive and easy to comprehend concerning
        homogeneity and completeness.
    - When you are uncertain about the cluster structure, since V-measure makes no assumptions about the
        cluster structure and is adaptable to all clustering algorithms.
    - When you want to establish a basis for comparison since homogeneity, completeness, and V-measure
        are bounded between [0, 1]. This bounded range simplifies score comparisons across various algorithms.
    Not Use:
    - When you don't have the ground truth labels since homogeneity, completeness, and V-measure are extrinsic
        measures and necessitate ground truth cluster assignments.
    - When your sample size is less than 1000 and the number of clusters is over 10 since V-measure doesn't
        adjust for chance. This implies that random labelling may not produce zero scores, especially when the
        number of clusters is large.

    Args:
        true_labels (List): True labels
        pred_labels (List): Predicted labels
        beta (float, optional): Ratio of weight attributed to homogeneity vs completeness.
            beta > 1, completeness is weighted more strongly in the calculation.
            beta < 1, homogeneity is weighted more strongly.
            Defaults to 1.0.

    Returns:
        Tuple[float, float, float]: homogeneity, completeness, v_measure
    """
    hs = homogeneity_score(true_labels, pred_labels)
    cs = completeness_score(true_labels, pred_labels)
    v = v_measure_score(true_labels, pred_labels, beta=beta)
    return hs, cs, v


def calculate_mutual_similarity(
    true_labels: List, pred_labels: List
) -> Tuple[float, float, float]:
    """
    The Mutual Information (MI), Normalized Mutual Information (NMI), and Adjusted Mutual Information (AMI)
    measures quantify the degree of agreement between cluster assignments.
    A larger score indicates greater similarity between the clusters.

    Use:
    - When you want to establish a basis for comparison.
    - When you want to compare MI, NMI, and AMI with an upper bound of 1.
    Not Use:
    - When you don't have the ground truth labels.

    Args:
        true_labels (List): True labels
        pred_labels (List): Predicted labels

    Returns:
        Tuple[float, float, float]: mutual_score, normalized_mutual_score, adjusted_mutual_score
    """
    mi = mutual_info_score(true_labels, pred_labels)
    nmi = normalized_mutual_info_score(
        true_labels, pred_labels, average_method="geometric"
    )
    ami = adjusted_mutual_info_score(true_labels, pred_labels)
    return mi, nmi, ami


def calculate_rand_index(true_labels: List, pred_labels: List) -> Tuple[float, float]:
    """Rand Index (RI, ARI) measures the similarity between the cluster assignments by making pair-wise comparisons.
    A higher score signifies higher similarity.

    Use:
    - When you want interpretability
    - When you are unsure about cluster structure
    - When you want a basis for comparison
    Not Use:
    - When you do not have the ground truth labels.

    Args:
        true_labels (List): True labels
        pred_labels (List): Predicted labels

    Returns:
        Tuple[float, float]: rank index and adjusted rank index metrics
    """
    ri = rand_score(true_labels, pred_labels)
    ari = adjusted_rand_score(true_labels, pred_labels)
    return ri, ari


def calculate_fm_score(true_labels: List, pred_labels: List) -> float:
    """The Fowlkes-Mallows scores quantify the accuracy of the cluster assignments using pairwise precision and recall.
    A higher score indicates a greater degree of similarity.

    Use:
    - When you are uncertain about the cluster structure since the Fowlkes-Mallows Score does not assume any specific
        cluster structure and can be used with any clustering algorithm.
    - When you need to establish a basis for comparison since the Fowlkes-Mallows Score has a maximum limit of 1.
        This restricted range simplifies the comparison of scores across various algorithms.
    Not Use:
    - When you don't have the ground truth labels since the Fowlkes-Mallows Scores are extrinsic measures that need
        ground truth cluster assignments.

    Args:
        true_labels (List): True labels
        pred_labels (List): Predicted labels

    Returns:
        float: Fowlkes Mallows score
    """
    return fowlkes_mallows_score(true_labels, pred_labels)



from collections import Counter
from typing import List


def preorder_solidity_traversal(
    node, depth: int = 2, chain_ratio: float = 4, parent_split_ratio: float = 10
) -> List:
    """
    The aim of this method is to return possible unique list of nominates for being the final nodes
    (pseudo-roots for stable solid clusters)
    """

    result = []
    """
    If a node has children, we first check if the chain.solidity is ratio higher than own solidity.
    If that is the case - the node is "weak" and we should nominate children as final nodes (result list).

    We recursively check if children's solidity is higher than parents. If such a case, we enter recursion
    and add the result to our results list.
    Contrary, if there are no children more solid than the parent - we add the current parent node.
    """
    if node.children:
        if node.chain_solidity > parent_split_ratio * node.solidity:
            for child in node.children:
                result += [child]

        add = True
        if node.children[0].solidity >= node.solidity:
            result += preorder_solidity_traversal(
                node.children[0],
                depth=depth,
                chain_ratio=chain_ratio,
                parent_split_ratio=parent_split_ratio,
            )
            add = False
        if node.children[1].solidity >= node.solidity:
            result += preorder_solidity_traversal(
                node.children[1],
                depth=depth,
                chain_ratio=chain_ratio,
                parent_split_ratio=parent_split_ratio,
            )
            add = False
        if add:
            result += [node]

    """
    There are usual cases, when *grand children are more solid than their *(grand)parents.
    For this case we implement depth parameter. depth = 2, allows you to check the depth of 2 generations,
    aka grandchildren.
    For *grandchildren we check the following conditions:
    a. grandchild.solidity and grandchild.chain_solidity / grandchild.solidity -- if the chain solidity of grandchild
        is high -- we continue the dig and add recursive result
    b. grandchild.solidity and grandchild.solidity >= node.solidity -- if the solidity of the grandchild is higher than
        the nodes' -- we continue the dig and add recursive result
    
    else the current node is added to the final result
    """
    if depth >= 1 and node.children:
        for child in node.children:
            if child.children:
                for grandchild in child.children:
                    if grandchild.solidity and grandchild.solidity >= node.solidity:
                        result += preorder_solidity_traversal(
                            grandchild, depth=depth - 1, chain_ratio=chain_ratio
                        )
                    else:
                        result += [node]

    # the end result might have nodes added by different conditions - so we return a unique list
    return list(set(result))


def bottom_up_cork_leaves(nodes: List, basic_leaves: set, node_tree, root_id: int):
    """
    The aim of this method is to find possible "corks" to make the whole tree full covered.
    TODO: check if a parent cork is better than current or if children is better than current cork
    TODO: define a condition for above TODO

    The idea is to check all the ancestors leaf-to-root for all basic_leafs (get_ancestors() method).
    If the root_id is present in the basic_ancestors set -- that means not all ground leaves are covered.
    The leaves_ancestors list is created to eliminate intermediate stable nodes and the corks dict is
    sorted by the frequency of node appearance and it's level (value).
    While corks object has nodes we add the top leaf to our nodes list, remove all descendants of the added
    cork and check the corks dict yet again.

    Args:
        nodes (List): _description_
        basic_leaves (set): _description_
        node_tree (_type_): _description_
        root_id (int): id of the root element

    Returns:
        _type_: _description_
    """
    root_id = node_tree.tree_structure[0].root.id if not root_id else root_id
    basic_ancestors = get_ancestors(basic_leaves, nodes)
    basic_ancestors_set = set([e for l in list(basic_ancestors.values()) for e in l])

    if root_id in basic_ancestors_set:
        leaves_ancestors = get_ancestors(nodes)
        leaves_ancestors_set = set(
            [e for l in list(leaves_ancestors.values()) for e in l]
            + list(leaves_ancestors.keys())
        )
        corks = {
            k: v
            for k, v in Counter(
                [e for l in list(basic_ancestors.values()) for e in l]
            ).items()
            if k not in leaves_ancestors_set
        }

        while corks:
            new_leaf = sorted(corks.items(), key=lambda x: (x[1], -x[0]), reverse=True)[
                0
            ][0]
            nodes.append(node_tree.tree_structure[new_leaf])
            descendants_list = [
                x.id for x in node_tree.tree_structure[new_leaf].descendants
            ] + [new_leaf]
            corks = {key: corks[key] for key in corks if key not in descendants_list}

    return nodes


# used when the number of clusters is specified
# maybe some minmax implementation to optimize for large number of clusters
def get_all_covered_combinations(node, n):
    """
    Get all possible combinations of covered nodes in a tree up to a given depth.

    Args:
        node (TreeNode): The root node of the tree.
        n (int): The depth up to which the combinations are to be generated.

    Returns:
        list: A list of lists containing all possible combinations of covered nodes.
    """
    result = [[node]]  # Initialize the result list with the root node

    for _ in range(1, n):
        new_result = []  # Create a new list to store the updated combinations
        for combination in result:
            for idx, node in enumerate(combination):
                if node.children:  # Check if the current node has children
                    # Generate a new combination by removing the current node and adding its children
                    res_list = (
                        combination[:idx] + combination[idx + 1 :] + list(node.children)
                    )
                    res_list = sorted(
                        res_list, key=lambda x: x.id
                    )  # Sort the nodes by their id
                    if (
                        res_list not in new_result
                    ):  # Check if the new combination is not already in the result list
                        new_result.append(
                            res_list
                        )  # Add the new combination to the result list
        result = new_result  # Update the result list with the new combinations

    return result


def get_max_solidity(combinations):
    """
    Get the combination of objects with the maximum sum of solidity.

    Args:
        combinations (list): A list of combinations, where each combination is a list of objects.

    Returns:
        list: The combination of objects with the maximum sum of solidity.
    """
    current_max = float(
        "-inf"
    )  # Initialize the maximum sum of solidity to a very small value
    best_combination = None  # Initialize the best combination to None

    for combination in combinations:
        _sum = sum(
            [x.solidity for x in combination]
        )  # Calculate the sum of solidity for the current combination
        if (
            _sum > current_max
        ):  # If the sum of solidity is greater than the current maximum
            current_max = _sum  # Update the current maximum
            best_combination = combination  # Update the best combination

    return best_combination  # Return the combination of objects with the maximum sum of solidity


def preorder_cork_traversal(
    node, visited: set = set(), min_leaf: int = 1, depth: int = 2
):
    """
    Recursively traverses a binary tree in a preorder manner and collects all nodes to visited
    that meet one of the following conditions:
    a. one of children's children_count attribute is less than min_leaves
    b. not (child_1.self_cover or child_1.chain_cover) or not (child_2.self_cover or child_2.chain_cover)
        -- at least one of the children is both not self_cover and chain_cover

    Args:
        node (_type_): a Node object, the root node of the subtree to traverse.
        visited (set): a set of Node objects, keeping track of visited nodes that do not meet certain conditions.
        min_leaf (int): an integer, the minimum number of children a node should have.
        depth (int): Maximum depth to check for solidity comparison.

    Returns:
        None, the function updates the visited set.
    """

    def is_root_solidity_higher(node, child=None, depth=1):
        """
        Check if root node's solidity is higher than its children's solidity up to a certain depth.

        Args:
            node (Node): Root node of the binary tree.
            depth (int): Maximum depth to check for solidity comparison.

        Returns:
            bool: True if root node's solidity is higher than its children's solidity up to a certain depth, False otherwise.
        """
        if not node.children or depth == 0:
            return True

        if child and not child.children:
            return True

        if not child:
            for child in node.children:
                if node.solidity <= child.solidity:
                    return False
                elif not is_root_solidity_higher(node, child, depth=depth - 1):
                    return False
        else:
            for grandchild in child.children:
                if node.solidity <= grandchild.solidity:
                    return False
                elif not is_root_solidity_higher(node, grandchild, depth=depth - 1):
                    return False

        return True

    if is_root_solidity_higher(node=node, child=None, depth=depth):
        visited.add(node)
        return

    if node.children:
        child_1, child_2 = node.children[0], node.children[1]
        if (
            child_1.children_count < min_leaf
            or child_2.children_count < min_leaf
            or not child_1.children
            or not child_2.children
        ):
            visited.add(node)
            return

        if not (child_1.self_cover or child_1.chain_cover) or not (
            child_2.self_cover or child_2.chain_cover
        ):
            visited.add(node)
            return

        preorder_cork_traversal(child_1, visited, min_leaf)
        preorder_cork_traversal(child_2, visited, min_leaf)


def calculate_chain_solidity(node):
    """
    Calculating the solidity for the tree recursively.
    Example: chain_solidity_calculation(root)

    Args:
        node (_type_): AnyTree Node

    Returns:
        _type_: AnyTree Node
    """
    if node.children:
        # Process the current node
        _sum = sum(child.chain_solidity for child in node.children)
        node.chain_solidity = (
            _sum if node.chain_solidity < _sum else node.chain_solidity
        )

        # Recursively process the children nodes
        for child in node.children:
            calculate_chain_solidity(child)

    return node


def postorder_chain_cover_traversal(node):
    """Modifies parent chain_cover attribute to True if both children's self_cover is True

    Args:
        node (_type_): _description_

    Returns:
        _type_: _description_
    """
    if node.children:
        res_1 = postorder_chain_cover_traversal(node.children[0])
        res_2 = postorder_chain_cover_traversal(node.children[1])

        if res_1 and res_2:  # or node.self_cover:
            node.chain_cover = True
            return True

    return node.self_cover


# ratio of solidities that are larger in depth = min depth
def postorder_children_count_traversal(node):
    if node.children:
        res_1 = postorder_children_count_traversal(node.children[0])
        res_2 = postorder_children_count_traversal(node.children[1])
        node.children_count = res_1 + res_2
    else:
        node.children_count = 1
    return node.children_count


# use Node.ancestors parameter?
def get_ancestors(leaves, check_list=[]):
    """
    For all leaves in leaves collect ancestors in an ancestors dictionary.
    If a particular leaf is present in the check_list, break the while loop.

    TODO: remove node = leaf

    Args:
        leaves (_type_): _description_
        check_list (list, optional): _description_. Defaults to [].

    Returns:
        _type_: _description_
    """
    ancestors = {}
    for leaf in leaves:
        node = leaf
        ancestors[leaf.id] = []
        while node.parent:
            ancestors[leaf.id] += [node.parent.id]
            if node.parent not in check_list:
                node = node.parent
            else:
                break
    return ancestors


# to consoder using for final_nodes
# Node.root
# Node.siblings



from matplotlib import pyplot as plt
import seaborn as sns
import numpy as np
import scipy.cluster.hierarchy as shc
from typing import List
from matplotlib.colors import ListedColormap
from bk_clustering.utilities import metrics
from anytree import RenderTree


def get_scatter(xarr: List, yarr: List, figsize=(16, 10), **kwargs):
    """Draw scatter plot

    Args:
        _xarr (List): x coordinates
        _yarr (List): y coordinates
        figsize (tuple, optional): _description_. Defaults to (16, 10).

    Returns:
        _type_: _description_
    """
    plt.figure(figsize=figsize)
    scatter = plt.scatter(xarr, yarr, **kwargs)
    plt.legend(*scatter.legend_elements())
    return plt


def get_heatmap(matrix, figsize=(15, 12)):
    """Draw heatmap

    Args:
        matrix (_type_): nxn list
        figsize (tuple, optional): _description_. Defaults to (15, 12).

    Returns:
        _type_: sns.heatmap
    """
    fig, ax = plt.subplots(figsize=figsize)
    return sns.heatmap(
        matrix,
        ax=ax,
        annot=True,
    )


def get_error_plots(_xarr: List, _yarr: List, true_labels: List, pred_labels: List):
    """Calculating accuracy of predicted labels.
    1. Map closest predicted cluster with ground true cluster
    2. Calculate accuracy by finding how many data points outside the max

    Args:
        _xarr (List): x coordinates
        _yarr (List): y coordinates
        true_labels (List): True labels
        pred_labels (List): Predicted labels

    Returns:
        Tuple[int, int, float]: number of correct predictions, number of predicted classes, accuracy
    """

    map_dict, correct_list = metrics.get_correct_list(true_labels, pred_labels)
    fig, axs = plt.subplots(2, 2, figsize=(15, 12))
    # define custom color map - need to handle 100% of fails
    cmap = ListedColormap([[1, 0, 0], [0, 1, 0]])
    if sum(correct_list) == len(_xarr):
        cmap = ListedColormap([[0, 1, 0], [1, 0, 0]])
    axs[0, 0].scatter(_xarr, _yarr)
    axs[0, 0].set_title("initial")
    axs[0, 1].scatter(_xarr, _yarr, c=true_labels)
    axs[0, 1].set_title("ground truth")
    axs[1, 0].scatter(_xarr, _yarr, c=pred_labels)
    axs[1, 0].set_title("proposed method")
    axs[1, 1].scatter(_xarr, _yarr, c=correct_list, cmap=cmap)
    axs[1, 1].set_title("error points")
    return axs


def augmented_dendrogram(
    dtf, limit=20, ids=[], positive_ids=[], log=False, figsize=(14, 10), *args, **kwargs
):
    if dtf.shape[1] > 4:
        _dtf = dtf.iloc[:, :4]
    else:
        _dtf = dtf.copy()
    if log == True:
        _dtf["distance"] = _dtf["distance"].apply(np.log)
    ddata = get_dendrogram(_dtf, figsize=figsize, **kwargs)
    h_mapping = {}
    if "point_id" in dtf.columns:
        h_mapping = {
            x[0]: x[1]
            for x in dtf.sort_values("distance", ascending=False)[
                ["distance", "point_id"]
            ]
            .iloc[:limit]
            .to_dict("split")["data"]
        }

    if not kwargs.get("no_plot", False):
        for idx, (i, d) in enumerate(zip(ddata["icoord"], ddata["dcoord"])):
            x = 0.5 * sum(i[1:3])
            y = d[1]
            if ids:
                if ids[idx] in positive_ids:
                    plt.plot(x, y, "go")
                else:
                    plt.plot(x, y, "ro")
            if y in h_mapping:
                plt.annotate(
                    h_mapping[y],
                    (x, y),
                    xytext=(10, 15),
                    textcoords="offset points",
                    va="top",
                    ha="center",
                )
    return ddata


def get_dendrogram(dtf, title=None, figsize=(14, 10), **kwargs):
    """Draw dendrogram

    Args:
        dtf (_type_): modified df with additional d parameter calculated
        figsize (tuple, optional): _description_. Defaults to (14, 10).

    Returns:
        _type_: _description_
    """
    plt.figure(figsize=figsize)
    plt.title(title)
    return shc.dendrogram(dtf, **kwargs)


def draw_tree(tree, maxlevel=5):
    """
    Draw a tree structure using ASCII art representation.

    Args:
        tree: Tree object
            The tree structure to be drawn.
        maxlevel: int, optional (default=5)
            The maximum level of the tree to be drawn.

    Returns:
        None
    """
    for pre, fill, node in RenderTree(tree, maxlevel=maxlevel):
        # Print the tree node's id, indented with pre and fill characters
        print("%s%s" % (pre, node.id))



from timeit import default_timer as timer
from typing import List
import numpy as np
from utilities import load_save, metrics, density_peak
from sklearn.cluster import (
    KMeans,
    MiniBatchKMeans,
    AffinityPropagation,
    AgglomerativeClustering,
    Birch,
    DBSCAN,
    MeanShift,
    OPTICS,
    SpectralClustering,
)
from sklearn.mixture import GaussianMixture
import hdbscan


def get_data(dataset_name: str):
    """
    Load and preprocess a dataset from an ARFF file.

    Args:
        dataset_name (tuple): A tuple containing the name of the ARFF file and the name of the class attribute.

    Returns:
        tuple: A tuple containing the feature matrix (X) and the true labels (true_labels).
    """
    # Load the ARFF file using the specified dataset_name
    df = load_save.read_arff(dataset_name[0], dataset_name[1])

    # Extract the feature matrix (X) and the true labels from the loaded ARFF file
    X, true_labels = df.loc[:, df.columns != "class"], df["class"]

    return X, true_labels


def collect_all_results(
    results: dict, dataset_name: str, true_labels: List, model, start
):
    """
    Collect and store the results of a model's performance on a dataset.

    Args:
        results (dict): A dictionary containing the results of multiple models on different datasets.
        dataset_name (str): The name of the dataset.
        true_labels (array-like): The true labels of the dataset.
        model (object): The trained model.
        start (float): The starting time when the model evaluation began.

    Returns:
        dict: The updated results dictionary with the collected performance metrics.
    """
    # Calculate the error metrics for the model's predicted labels
    error_results = metrics.calculate_metrics(true_labels, model.labels_)

    # Format the error results and store them in the results dictionary along with the elapsed time
    results[dataset_name] = load_save.format_results(error_results, timer() - start)

    return results


# TODO: itterate kmeans through different random seeds and check if some our outliered?
# TODO: check how the results change with number of clusters mistaken
def run_kmeans(dataset_names: List, number_of_clusters: List, random_state: int = 1):
    """
    Run the K-means clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.
        random_state (int): An optional random seed for reproducibility. Default is 0.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = KMeans(n_clusters=n_clusters, random_state=random_state).fit(
            X
        )  # Train the K-means model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = (
        f"./../results/k_means_results.json"  # Define the filename to save the results
    )
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_kmeans_mini_batch(
    dataset_names: List, number_of_clusters: List, random_state: int = 1
):
    """
    Run the Mini-Batch K-means clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.
        random_state (int): An optional random seed for reproducibility. Default is 0.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = MiniBatchKMeans(n_clusters=n_clusters, random_state=random_state).fit(
            X
        )  # Train the Mini-Batch K-means model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = f"./../results/mini_batch_kmeans_results.json"  # Define the filename to save the results
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_affinity_propagation(
    dataset_names: List, number_of_clusters: List, random_state: int = 1
):
    """
    Run the Affinity Propagation clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.
        random_state (int): An optional random seed for reproducibility. Default is 0.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = AffinityPropagation(random_state=random_state).fit(
            X
        )  # Train the Affinity Propagation model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = (
        f"./../results/affinity_results.json"  # Define the filename to save the results
    )
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_agglomerative(dataset_names: List, number_of_clusters: List):
    """
    Run the Agglomerative Clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = AgglomerativeClustering(n_clusters=n_clusters).fit(
            X
        )  # Train the Agglomerative Clustering model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = f"./../results/agglomerative_results.json"  # Define the filename to save the results
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_birch(dataset_names, number_of_clusters):
    """
    Run the Birch Clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = Birch(n_clusters=n_clusters).fit(X)  # Train the Birch Clustering model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = (
        f"./../results/birch_results.json"  # Define the filename to save the results
    )
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_dbscan(dataset_names: List, number_of_clusters: List):
    """
    Run the DBSCAN clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = DBSCAN().fit(X)  # Train the DBSCAN model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = (
        f"./../results/dbscan_results.json"  # Define the filename to save the results
    )
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_mean_shift(dataset_names: List, number_of_clusters: List):
    """
    Run the Mean Shift clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = MeanShift().fit(X)  # Train the Mean Shift model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = f"./../results/mean_shift_results.json"  # Define the filename to save the results
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_optics(dataset_names: List, number_of_clusters: List):
    """
    Run the OPTICS clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = OPTICS().fit(X)  # Train the OPTICS model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = (
        f"./../results/optics_results.json"  # Define the filename to save the results
    )
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_spectral(dataset_names: List, number_of_clusters: List):
    """
    Run the Spectral Clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        model = SpectralClustering(n_clusters=n_clusters).fit(
            X
        )  # Train the Spectral Clustering model
        results = collect_all_results(
            results, dataset_name, true_labels, model, start
        )  # Collect and store results

    filename = (
        f"./../results/spectral_results.json"  # Define the filename to save the results
    )
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_gmm(dataset_names: List, number_of_clusters: List, random_state=1):
    """
    Run the Gaussian Mixture Model (GMM) clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.
        random_state (int): An optional random seed for reproducibility. Default is 1.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        # as it was discovered for some datasets ("jm1", "KDDTest+", etc.) GaussianMixture throws an error
        try:
            start = timer()  # Record the start time of model evaluation
            X, true_labels = get_data(
                dataset_name
            )  # Load the dataset and extract true labels
            model = GaussianMixture(
                n_components=n_clusters, random_state=random_state
            ).fit(
                X
            )  # Train the GMM model
            pred_labels = model.predict(X)  # Predict the cluster labels
            error_results = metrics.calculate_metrics(
                true_labels, pred_labels
            )  # Calculate error metrics
            results[dataset_name] = load_save.format_results(
                error_results, timer() - start
            )  # Collect and store results
        except Exception as e:
            print(e)  # Print any exception that occurs during model evaluation

    filename = f"./../results/gaussian_mixture_results.json"  # Define the filename to save the results
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_density_peak(dataset_names: List, number_of_clusters: List):
    """
    Run the Density Peak clustering algorithm on multiple datasets and collect and save the results.

    Args:
        dataset_names (list): A list of dataset names.
        number_of_clusters (list): A list of the number of clusters to use for each dataset.

    Returns:
        None
    """
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        start = timer()  # Record the start time of model evaluation
        X, true_labels = get_data(
            dataset_name
        )  # Load the dataset and extract true labels
        pred_labels = density_peak.run_density_peak(
            X
        )  # Run the Density Peak algorithm to get predicted labels
        error_results = metrics.calculate_metrics(
            true_labels, pred_labels
        )  # Calculate clustering metrics
        results[dataset_name] = load_save.format_results(
            error_results, timer() - start
        )  # Collect and store results

    filename = f"./../results/density_peak_results.json"  # Define the filename to save the results
    load_save.save_json(results, filename)  # Save the results to a JSON file


def run_hdbscan(dataset_names: List, number_of_clusters: List, random_state=1):
    results = {}  # Initialize an empty dictionary to store the results

    # Loop through each dataset and corresponding number of clusters
    for dataset_name, n_clusters in zip(dataset_names, number_of_clusters):
        try:
            start = timer()  # Record the start time of model evaluation
            X, true_labels = get_data(
                dataset_name
            )  # Load the dataset and extract true labels

            model = hdbscan.HDBSCAN()
            model.fit(X)  # Train the hdbscan model
            pred_labels = model.labels_  # Cluster labels
            error_results = metrics.calculate_metrics(
                true_labels, pred_labels
            )  # Calculate error metrics
            results[dataset_name] = load_save.format_results(
                error_results, timer() - start
            )  # Collect and store results
        except Exception as e:
            print(e)  # Print any exception that occurs during model evaluation

    filename = (
        f"./../results/hdbscan_results.json"  # Define the filename to save the results
    )
    load_save.save_json(results, filename)  # Save the results to a JSON file


def generate_corr_matrix(dataset_size_dict):
    import itertools
    from collections import OrderedDict
    from bk_clustering import postprocessing, metrics, load_save

    filename = "./../results/predictions/bk_clustering_predictions.json"
    bk_clustering_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/k_means_predictions.json"
    k_means_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/mini_batch_kmeans_predictions.json"
    mini_batch_kmeans_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/affinity_predictions.json"
    affinity_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/agglomerative_predictions.json"
    agglomerative_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/birch_predictions.json"
    birch_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/hdbscan_predictions.json"
    hdbscan_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/dbscan_predictions.json"
    dbscan_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/mean_shift_predictions.json"
    mean_shift_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/optics_predictions.json"
    optics_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/gaussian_mixture_predictions.json"
    gaussian_mixture_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    filename = "./../results/predictions/density_peak_predictions.json"
    density_peak_predictions = {
        eval(k): v
        for (k, v) in OrderedDict(sorted(load_save.load_json(filename).items())).items()
    }

    prediction_dict = {
        "bk_clustering": [
            postprocessing.convert_to_array(eval(x))
            for x in bk_clustering_predictions.values()
        ],
        "kmeans": [eval(x) for x in k_means_predictions.values()],
        "affinity": [eval(x) for x in affinity_predictions.values()],
        "dbscan": [eval(x) for x in dbscan_predictions.values()],
        "hdbscan": [eval(x) for x in hdbscan_predictions.values()],
        "density_peak": [eval(x) for x in density_peak_predictions.values()],
        "mean_shift_predictions": [eval(x) for x in mean_shift_predictions.values()],
        "optics": [eval(x) for x in optics_predictions.values()],
        "agglomerative": [eval(x) for x in agglomerative_predictions.values()],
        "birch": [eval(x) for x in birch_predictions.values()],
        "gmm": [eval(x) for x in gaussian_mixture_predictions.values()],
        "mb_kmeans": [eval(x) for x in mini_batch_kmeans_predictions.values()],
    }

    raw_metrics_dict = {}
    for key in list(itertools.combinations(prediction_dict, 2)):
        raw_metrics_dict[key] = {}
        raw_metrics_dict[key]["v_measure"] = np.array(
            [
                list(
                    metrics.calculate_v_measure(ds1, ds2)
                    for ds1, ds2 in zip(
                        prediction_dict[key[0]], prediction_dict[key[1]]
                    )
                )
            ]
        )
        raw_metrics_dict[key]["mutual_similarity"] = np.array(
            [
                list(
                    metrics.calculate_mutual_similarity(ds1, ds2)
                    for ds1, ds2 in zip(
                        prediction_dict[key[0]], prediction_dict[key[1]]
                    )
                )
            ]
        )
        raw_metrics_dict[key]["rand_index"] = np.array(
            [
                list(
                    metrics.calculate_rand_index(ds1, ds2)
                    for ds1, ds2 in zip(
                        prediction_dict[key[0]], prediction_dict[key[1]]
                    )
                )
            ]
        )
        raw_metrics_dict[key]["fm_score"] = np.array(
            [
                list(
                    metrics.calculate_fm_score(ds1, ds2)
                    for ds1, ds2 in zip(
                        prediction_dict[key[0]], prediction_dict[key[1]]
                    )
                )
            ]
        )

    final_metrics_dict = {}
    for key in list(itertools.combinations(prediction_dict, 2)):
        final_metrics_dict[key] = {}
        final_metrics_dict[key]["v_measure_0"] = raw_metrics_dict[key]["v_measure"][
            :, :, 0
        ][0]
        final_metrics_dict[key]["v_measure_1"] = raw_metrics_dict[key]["v_measure"][
            :, :, 1
        ][0]
        final_metrics_dict[key]["v_measure_2"] = raw_metrics_dict[key]["v_measure"][
            :, :, 2
        ][0]
        final_metrics_dict[key]["mutual_similarity_0"] = raw_metrics_dict[key][
            "mutual_similarity"
        ][:, :, 0][0]
        final_metrics_dict[key]["mutual_similarity_1"] = raw_metrics_dict[key][
            "mutual_similarity"
        ][:, :, 1][0]
        final_metrics_dict[key]["mutual_similarity_2"] = raw_metrics_dict[key][
            "mutual_similarity"
        ][:, :, 2][0]
        final_metrics_dict[key]["rand_index_0"] = raw_metrics_dict[key]["rand_index"][
            :, :, 0
        ][0]
        final_metrics_dict[key]["rand_index_1"] = raw_metrics_dict[key]["rand_index"][
            :, :, 1
        ][0]
        final_metrics_dict[key]["fm_score"] = raw_metrics_dict[key]["fm_score"][:][0]

    json_dict = {k: {} for k in final_metrics_dict}
    for k in final_metrics_dict:
        for m in final_metrics_dict[k]:
            json_dict[k][m] = load_save.encode_json(
                {
                    dataset: metric
                    for (dataset, metric) in zip(
                        dataset_size_dict, final_metrics_dict[k][m]
                    )
                }
            )

    load_save.save_json(json_dict, "./../results/aggregations/aggregated.json")



from scipy.cluster.hierarchy import linkage
from collections import Counter
import pandas as pd
from timeit import default_timer as timer

import sys

# TODO: think of a better way
sys.setrecursionlimit(25000)

# TODO: confidence of the split through sum of solidity/number of points
# TODO: Find dataset with ground truth for mulitlabelling?

from bk_clustering.utilities import (
    calculation_utilities,
    cluster_calculations,
    tree_traversal,
    load_save,
    plot_utilities,
    postprocessing,
)


class BurjKhalifaClustering:
    def __init__(
        self,
        dH0: float = 0,
        depth: int = 2,
        chain_ratio: float = 5,
        parent_split_ratio: float = 10,
        min_leaves: int = 0,
        n_clusters: int = None,
        linkage=None,
    ):
        self.n_clusters = n_clusters
        self.dH0 = dH0
        self.depth = depth
        self.chain_ratio = chain_ratio
        self.parent_split_ratio = parent_split_ratio
        self.min_leaves = min_leaves
        self.linkage = linkage
        self.linkage_types = ["ward", "single"]

    def fit(
        self,
        X,
    ):
        """
        Wrapper function for clustering algorithm.

        Args:
            X (array-like): Input data for clustering.
            linkage_type (str, optional): Name of the clustering method. Defaults to None.
            min_depth (int, optional): Minimum depth for clustering. Defaults to 2.
            chain_ratio (int, optional): Ratio for chain merge threshold. Defaults to 5.
            parent_split_ratio (int, optional): Ratio for parent split threshold. Defaults to 10.
            min_leaf_ratio (float, optional): Minimum leaf ratio for clustering. Defaults to 0.01.
            number_of_clusters (int, optional): Number of clusters to form. Defaults to None.

        Returns:
            tuple: A tuple containing result_dict, dtf, cluster_df, cluster_info, and node_tree.
        """

        # If linkage_type is not provided, calculate clusters for "ward" and "single" methods
        if not self.linkage:
            method_dict = {}
            for linkage in self.linkage_types:
                method_dict[linkage] = {}
                (
                    result_dict,
                    dtf,
                    cluster_df,
                    cluster_info,
                    node_tree,
                ) = self.calculate_clusters(
                    X,
                    linkage,
                )
                if result_dict:
                    method_dict[linkage]["result_dict"] = result_dict
                    method_dict[linkage]["dtf"] = dtf
                    method_dict[linkage]["cluster_df"] = cluster_df
                    method_dict[linkage]["cluster_info"] = cluster_info
                    method_dict[linkage]["node_tree"] = node_tree
                    method_dict[linkage]["average_solidity"] = sum(
                        [
                            (
                                Counter(result_dict.values())[x]
                                * cluster_info[x]["solidity"]
                            )
                            for x in set(result_dict.values())
                        ]
                    )
                else:
                    method_dict[linkage]["average_solidity"] = float("-inf")

            # Select method with maximum average_solidity
            linkage = max(method_dict.items(), key=lambda k: k[1]["average_solidity"])[
                0
            ]
            result_dict = method_dict[linkage]["result_dict"]
            dtf = method_dict[linkage]["dtf"]
            cluster_df = method_dict[linkage]["cluster_df"]
            cluster_info = method_dict[linkage]["cluster_info"]
            node_tree = method_dict[linkage]["node_tree"]

            self.linkage = linkage
        # If linkage_type is provided, calculate clusters for the given linkage_type
        else:
            (
                result_dict,
                dtf,
                cluster_df,
                cluster_info,
                node_tree,
            ) = self.calculate_clusters(X, linkage_type=self.linkage)

        self.labels_ = postprocessing.convert_to_array(result_dict.values())
        self.n_clusters = len(set(self.labels_))
        self.dtf_ = dtf
        self.cluster_info_ = cluster_info
        self.cluster_df_ = cluster_df
        self.node_tree_ = node_tree

        return self

    def calculate_clusters(self, X, linkage_type: str):
        """
        #TODO:
        # improve performance to calculate the distance matrix once

        import scipy.spatial.distance as distance
        from . import _hierarchy

        replace with
        n = int(distance.num_obs_y(y))
        method_code = _LINKAGE_METHODS[method]

        if method == 'single':
            result = _hierarchy.mst_single_linkage(y, n)
        elif method in ['complete', 'average', 'weighted', 'ward']:
            result = _hierarchy.nn_chain(y, n, method_code)
        """
        # calculating linkage dataframe
        linkage_df = pd.DataFrame(linkage(X, linkage_type))

        # renaming columns, assigning types
        linkage_df.columns = ["point_1", "point_2", "distance", "points_count"]
        linkage_df[["point_1", "point_2"]] = linkage_df[["point_1", "point_2"]].astype(
            int
        )
        linkage_df["point_id"] = (
            linkage_df.index + linkage_df["points_count"].iloc[-1]
        ).astype(int)

        # defining root_id
        root_id = linkage_df["point_id"].iloc[-1]
        min_leaves = self.min_leaves if self.min_leaves else 1

        # normalizing the distance parameter
        linkage_df["distance"] = calculation_utilities.normalize(linkage_df["distance"])

        # formal density as ratio of the number of data points per distance
        # linkage matrix with formal density calculated as number of data_point/normalized_distance
        linkage_df["formal_density"] = (
            linkage_df["points_count"] / linkage_df["distance"]
        )

        # building a ClusterTree
        node_tree = cluster_calculations.ClusterTree(linkage_df)
        node_tree.build_tree()
        root = node_tree.tree_structure[root_id]

        # creating a ClusterInfoObject object and defining cluster_info
        cluster_info_object = cluster_calculations.ClusterInfoObject(
            linkage_df["points_count"].iloc[-1], linkage_df, self.dH0
        )
        cluster_info_object.set_cluster_info(node_tree.tree_structure)
        cluster_info = cluster_info_object.cluster_info

        # setting leaf_nodes varaible
        leaf_nodes = set(
            cluster_info[x]["node"]
            for x in cluster_info
            if cluster_info[x]["level"] == 0
        )

        # setting node_tree parameters for every node: solidity, chain_solidity, self_cover, chain_cover and children_count
        for n in node_tree.tree_structure:
            node_tree.tree_structure[n].solidity = cluster_info[n]["solidity"]
            node_tree.tree_structure[n].chain_solidity = cluster_info[n]["solidity"]
            node_tree.tree_structure[n].self_cover = False
            node_tree.tree_structure[n].chain_cover = False

        if self.n_clusters:
            nodes_combinations = tree_traversal.get_all_covered_combinations(
                root, self.n_clusters
            )
            pre_cork_leaves = tree_traversal.get_max_solidity(nodes_combinations)

        else:
            tree_traversal.postorder_children_count_traversal(root)

            # update chain_solidity parameter
            tree_traversal.calculate_chain_solidity(node_tree.tree_structure[root_id])

            # get leaves calculated through preorder traversal
            leaves = tree_traversal.preorder_solidity_traversal(
                node_tree.tree_structure[root_id],
                depth=self.depth,
                parent_split_ratio=self.parent_split_ratio,
                chain_ratio=self.chain_ratio,
            )

            # remove leaves that don't match min_leaves condition
            leaves = [leaf for leaf in leaves if leaf.children_count >= min_leaves]

            # find missing corks
            leaves = tree_traversal.bottom_up_cork_leaves(
                leaves, leaf_nodes, node_tree, root_id
            )

            # modify the self_cover attribute for all leaves
            for node in leaves:
                node_tree.tree_structure[node.id].self_cover = True

            # modification of chain_cover attribute
            tree_traversal.postorder_chain_cover_traversal(root)

            # final stage of the end leaves selection
            pre_cork_leaves = set()
            tree_traversal.preorder_cork_traversal(
                root, pre_cork_leaves, min_leaves, self.depth
            )

        cork_nodes = tree_traversal.get_ancestors(leaf_nodes, pre_cork_leaves)

        # getting result as a dictionary {leaf_id:cluster_id}
        result_dict = {i: cork_nodes[i][-1] for i in sorted(list(cork_nodes.keys()))}

        # aggregating all information from the cluster_info object to a pandas dataframe
        cluster_df = pd.DataFrame.from_dict(cluster_info, orient="index")

        return result_dict, linkage_df, cluster_df, cluster_info, node_tree



# explicit function to normalize array
import pandas as pd
import scipy.spatial.distance as distance
import numpy as np


def normalize(arr: pd.Series, t_min: int = 1, t_max: int = 1000) -> pd.Series:
    """Normalize a pandas series to a custom range.

    Args:
        arr (pd.Series): The pandas series to normalize.
        t_min (int): The minimum value of the output range. Defaults to 1.
        t_max (int): The maximum value of the output range. Defaults to 1000.

    Returns:
        pd.Series: The normalized pandas series.

    Raises:
        ValueError: If the input series is empty or the output range is invalid.

    Example:
        >>> series = pd.Series([10, 20, 30, 40, 50])
        >>> normalize(series, 1, 1000)
        [1.0, 250.75, 500.5, 750.25, 1000.0]
    """
    if arr.empty:
        raise ValueError("Input series is empty.")
    if t_min >= t_max:
        raise ValueError("Invalid output range.")

    diff_arr = max(arr) - min(arr)
    if diff_arr:
        diff = t_max - t_min
        normalized_arr = [((i - min(arr)) * diff / diff_arr) + t_min for i in arr]
        return pd.Series(normalized_arr)
    return arr


def get_sq_matrix(data_points):
    """
    Calculates the square matrix of pairwise distances between data points.

    Args:
        data_points (array-like): Input data points as an array-like object.

    Returns:
        np.ndarray: Square matrix of pairwise distances between data points.
    """
    return distance.cdist(data_points, data_points)


def get_distance_matrix(xarr, yarr):
    """
    Calculates the distance matrix and the number of data points.

    Args:
        xarr (array-like): X-coordinates of data points.
        yarr (array-like): Y-coordinates of data points.

    Returns:
        tuple: Tuple containing the distance matrix and the number of data points.
    """
    data_points = np.vstack((xarr, yarr)).T
    sq_matrix = distance.cdist(data_points, data_points)
    return distance.squareform(sq_matrix), len(sq_matrix)



import numpy as np
import arff
import pandas as pd
from scipy.io import arff as scipy_arff
from bk_clustering.utilities import preprocessing
import ujson


def read_arff(
    folder, filename, skip_columns=[], miss_thresh_hard=5, miss_thresh_rel=0.05
):
    """
    Read an ARFF file from a specified folder and filename, and preprocess the data.

    Args:
        folder (str): The folder containing the ARFF file.
        filename (str): The filename of the ARFF file without the extension.

    Returns:
        pd.DataFrame: The preprocessed data as a pandas DataFrame, or None if the file is not found.
    """
    try:
        # Load ARFF data using scipy_arff and create a pandas DataFrame
        data = scipy_arff.loadarff(f"../data/{folder}/{filename}.arff")
        df = pd.DataFrame(data[0])
    except NotImplementedError:
        data = []
        for x in arff.load(f"../data/{folder}/{filename}.arff"):
            data.append(x._data)
        df = pd.DataFrame(data)
        df = df[[x for x in df.columns if type(x) == str]]
        save_column_list = []
        for col in df.columns:
            if df[df[col] == ""][col].count() and (
                df[df[col] == ""][col].count() <= miss_thresh_hard
                or df[df[col] == ""][col].count() <= miss_thresh_rel * df.shape[0]
            ):
                save_column_list.append(col)
        new_idx = df[df[save_column_list] != ""][save_column_list].dropna(axis=0).index
        if save_column_list:
            df = df.loc[new_idx]
        df = df[df != ""].dropna(axis=1)

    except FileNotFoundError:
        print(f"Dataset {folder}/{filename} not present")
        return None  # Return None if the file is not found
    df.columns = df.columns.str.lower()  # Convert column names to lowercase
    for col in skip_columns:
        if col in df.columns:
            df.drop(col, axis=1, inplace=True)
    df = preprocessing.preprocessing(
        df
    )  # Preprocess the data using a preprocessing function
    df = df.dropna(axis=1)
    return df  # Return the preprocessed data as a DataFrame


'''
def read_dataset(filename: str, train=True, validation=True):
    """
    filename -- dataset name. Example: a1,a2,s1,etc
    """
    xarr, yarr, validation_list = None, None, None
    relative_dir = "/"
    if train:
        f = open(f".{relative_dir}data/benchmarks/data/{filename}.txt", "r")
        data_points = np.array(
            [[float(_dp) for _dp in x.split(" ") if len(_dp) > 0] for x in f]
        )
        xarr = data_points[:, 0]
        yarr = data_points[:, 1]
    if validation:
        f = open(f".{relative_dir}data/benchmarks/validation/{filename}.pa", "r")
        validation_list = [int(x.replace("\n", "")) for x in f]
    return xarr, yarr, validation_list
'''


def save_json(data, filename):
    """
    Save data to a JSON file.

    Args:
        data (dict): The data to be saved as JSON.
        filename (str): The filename of the JSON file to be created.

    Returns:
        None
    """
    with open(filename, "w") as outfile:
        ujson.dump(data, outfile)  # Use ujson.dump to write data to the JSON file


def encode_json(data):
    return ujson.encode(data)


def load_json(filename):
    with open(filename) as outfile:
        data = ujson.load(outfile)
        return data


def format_results(results: dict, time: float) -> dict:
    """
    Format the results dictionary by adding the performance time.

    Args:
        results (dict): The results dictionary.
        time (float): The performance time to be added to the results.

    Returns:
        dict: The formatted results dictionary with the performance time added.
    """
    obj = {**results}  # Create a shallow copy of the results dictionary
    obj["performance_s"] = time  # Add the performance time to the results
    return obj



from anytree import AnyNode


class ClusterTree:
    def __init__(self, dtf):
        self.dtf = dtf

    def build_tree(self):
        self.tree_structure = {}
        for i, row in self.dtf.sort_values(by="point_id", ascending=False)[
            ["point_1", "point_2", "point_id"]
        ].iterrows():
            p1, p2, pi = row.values
            self.tree_structure[pi] = (
                AnyNode(id=pi)
                if pi not in self.tree_structure
                else self.tree_structure[pi]
            )
            self.tree_structure[p1] = AnyNode(id=p1, parent=self.tree_structure[pi])
            self.tree_structure[p2] = AnyNode(id=p2, parent=self.tree_structure[pi])

    # def add_property(self):


class ClusterInfoObject:
    """
    Class to store node data for clustering algorithm
    """

    def __init__(self, number_of_points, dtf, dH0):
        """
        Initializes the ClusterInfo object and populates it with data
        """
        self.number_of_points = number_of_points
        self.dtf = dtf
        self.dH0 = dH0

        # calculating maximum number of nodes in out tree: initial_nodes + (initial_nodes - 1)
        self.max_number_of_nodes = int(self.number_of_points * 2 - 1)

        # average formal density for bottom leafs
        self.density_0 = self.dtf.loc[
            self.dtf["points_count"] == 2, "formal_density"
        ].mean()

        # creating cluster_info dictionary
        self.cluster_info = {i: {} for i in range(self.max_number_of_nodes)}

    def set_cluster_info(self, tree):
        # setting basic node properties
        for i in self.cluster_info:
            # if node is initial data point
            if i < self.number_of_points:
                self.cluster_info[i]["height"] = 1
                self.cluster_info[i]["count"] = 1
                self.cluster_info[i]["density"] = self.density_0

                self.cluster_info[i]["children"] = []
            else:
                adjusted_i = int(i - self.number_of_points)

                self.cluster_info[i]["height"] = self.dtf.iloc[adjusted_i]["distance"]
                self.cluster_info[i]["count"] = self.dtf.iloc[adjusted_i][
                    "points_count"
                ]
                self.cluster_info[i]["density"] = self.dtf.iloc[adjusted_i][
                    "formal_density"
                ]
                # setting children property of current node
                self.cluster_info[i]["children"] = [
                    int(self.dtf.iloc[adjusted_i]["point_1"]),
                    int(self.dtf.iloc[adjusted_i]["point_2"]),
                ]

        for i in self.cluster_info:
            # setting nodes
            self.cluster_info[i]["node"] = tree[i]

            # setting hierarchical level
            self.cluster_info[i]["level"] = self.cluster_info[i]["node"].height

            # setting parent properties
            self.cluster_info[i]["parent"] = self.get_parent(i)

            # setting sibling properties
            # self.cluster_info[i]["siblings"] = self.get_siblings(i)

            # calculating height to next level
            self.cluster_info[i]["height_to_next"] = self.calculate_height_to_parent(i)

            # calculating solidity of the subcluster as density * height_to_next / height
            self.cluster_info[i]["solidity"] = (
                self.cluster_info[i]["density"]
                * self.cluster_info[i]["height_to_next"]
                / self.cluster_info[i]["height"]
            )

    # TODO remove
    def calculate_d_solidity(self, node):
        """
        Calculates the d_solidity value for a given node
        """
        _arr = [
            self.cluster_info[n]["solidity"]
            for n in self.cluster_info[node]["children"]
        ]
        return self.cluster_info[node]["solidity"] - sum(_arr) / 1

    # TODO remove
    def calculate_d_density(self, node):
        """
        Calculates the d_density value for a given node
        """
        _arr = [
            self.cluster_info[n]["density"] for n in self.cluster_info[node]["children"]
        ]
        return self.cluster_info[node]["density"] - sum(_arr) / 1

    def calculate_height_to_parent(self, node):
        """
        Calculating height to parent
        """
        parent = self.cluster_info[node]["parent"]
        return (
            self.cluster_info.get(parent, {"height": 0})["height"]
            - self.cluster_info[node]["height"]
        )

    def get_parent(self, node):
        """
        Get parent name
        """
        return (
            self.cluster_info[node]["node"].parent.id
            if self.cluster_info[node]["node"].parent
            else "root"
        )

    # replace with siblings attribute in anytree
    def get_siblings(self, node):
        """
        Get sibling.
        """
        if self.cluster_info[node]["parent"] != "root":
            return [
                x
                for x in self.cluster_info[self.cluster_info[node]["parent"]][
                    "children"
                ]
                if x != node
            ]
        else:
            return None



# code from https://github.com/lanbing510/DensityPeakCluster

import logging
import math
import scipy.spatial.distance as distance
import numpy as np

logger = logging.getLogger("dpc_cluster")


def load_paperdata(X):
    distances = distance.cdist(X, X)
    return distances, distances.max(), distances.min(), len(distances)


def select_dc(max_id, max_dis, min_dis, distances, auto=False):
    """
    Select the local density threshold, default is the method used in paper, auto is `autoselect_dc`

    Args:
        max_id    : max continues id
        max_dis   : max distance for all points
        min_dis   : min distance for all points
        distances : distance dict
        auto      : use auto dc select or not

    Returns:
        dc that local density threshold
    """
    logger.info("PROGRESS: select dc")
    if auto:
        return autoselect_dc(max_id, max_dis, min_dis, distances)
    percent = 2.0
    position = int(max_id * (max_id + 1) / 2 * percent / 100)
    dc = sorted(distances.flatten())[position * 2 + max_id]
    logger.info("PROGRESS: dc - " + str(dc))
    return dc


def autoselect_dc(max_id, max_dis, min_dis, distances):
    """
    Auto select the local density threshold that let average neighbor is 1-2 percent of all nodes.

    Args:
        max_id    : max continues id
        max_dis   : max distance for all points
        min_dis   : min distance for all points
        distances : distance dict

    Returns:
        dc that local density threshold
    """
    dc = (max_dis + min_dis) / 2

    while True:
        nneighs = sum([1 for v in distances if v < dc]) / max_id**2
        if nneighs >= 0.01 and nneighs <= 0.002:
            break
        # binary search
        if nneighs < 0.01:
            min_dis = dc
        else:
            max_dis = dc
        dc = (max_dis + min_dis) / 2
        if max_dis - min_dis < 0.0001:
            break
    return dc


def local_density(max_id, distances, dc, guass=True, cutoff=False):
    """
    Compute all points' local density

    Args:
        max_id    : max continues id
        distances : distance dict
        gauss     : use guass func or not(can't use together with cutoff)
        cutoff    : use cutoff func or not(can't use together with guass)

    Returns:
        local density vector that index is the point index that start from 1
    """
    assert guass and cutoff == False and guass or cutoff == True
    logger.info("PROGRESS: compute local density")
    guass_func = lambda dij, dc: math.exp(-((dij / dc) ** 2))
    cutoff_func = lambda dij, dc: 1 if dij < dc else 0
    func = guass and guass_func or cutoff_func
    rho = [-1] + [0] * max_id
    for i in range(0, max_id):
        for j in range(i + 1, max_id):
            rho[i + 1] += func(distances[i, j], dc)
            rho[j + 1] += func(distances[i, j], dc)
        if i % (max_id / 10) == 0:
            logger.info("PROGRESS: at index #%i" % (i))
    return np.array(rho, np.float32)


def min_distance(max_id, max_dis, distances, rho):
    """
    Compute all points' min distance to the higher local density point(which is the nearest neighbor)

    Args:
        max_id    : max continues id
        max_dis   : max distance for all points
        distances : distance dict
        rho       : local density vector that index is the point index that start from 1

    Returns:
        min_distance vector, nearest neighbor vector
    """
    logger.info("PROGRESS: compute min distance to nearest higher density neigh")
    sort_rho_idx = np.argsort(-rho)

    delta, nneigh = [0.0] + [float(max_dis)] * (len(rho) - 1), [0] * len(rho)
    delta[sort_rho_idx[0]] = -1.0

    for i in range(1, max_id):
        for j in range(0, i):
            old_i, old_j = sort_rho_idx[i], sort_rho_idx[j]
            if distances[old_i - 1, old_j - 1] < delta[old_i]:
                delta[old_i] = distances[old_i - 1, old_j - 1]
                nneigh[old_i] = old_j
        if i % (max_id / 10) == 0:
            logger.info("PROGRESS: at index #%i" % (i))
    delta[sort_rho_idx[0]] = max(delta)
    return np.array(delta, np.float32), np.array(nneigh, np.float32)


def calculate_clusters(
    X, density_threshold=1, distance_threshold=1, dc=None, auto_select_dc=False
):
    """
    Cluster the data

    Args:
        load_func          : load func to load data
        distance_f         : distance data file
        dc                 : local density threshold, call select_dc if dc is None
        density_threshold  : local density threshold for choosing cluster center
        distance_threshold : min distance threshold for choosing cluster center
        autoselect_dc      : auto select dc or not

    Returns:
        local density vector, min_distance vector, nearest neighbor vector
    """
    distances, max_dis, min_dis, max_id = load_paperdata(X)
    if dc == None:
        dc = select_dc(max_id, max_dis, min_dis, distances, auto=auto_select_dc)
    rho = local_density(max_id, distances, dc)
    delta, nneigh = min_distance(max_id, max_dis, distances, rho)
    logger.info("PROGRESS: start cluster")
    cluster, ccenter = {}, {}  # cl/icl in cluster_dp.m

    for idx, (ldensity, mdistance, nneigh_item) in enumerate(zip(rho, delta, nneigh)):
        if idx == 0:
            continue
        if ldensity >= density_threshold and mdistance >= distance_threshold:
            ccenter[idx] = idx
            cluster[idx] = idx
        else:
            cluster[idx] = -1

    # assignation
    ordrho = np.argsort(-rho)
    for i in range(ordrho.shape[0] - 1):
        if ordrho[i] == 0:
            continue
        if cluster[ordrho[i]] == -1 and nneigh[ordrho[i]]:
            cluster[ordrho[i]] = cluster[nneigh[ordrho[i]]]
        if i % (max_id / 10) == 0:
            logger.info("PROGRESS: at index #%i" % (i))

    # halo
    halo, bord_rho = {}, {}
    for i in range(1, ordrho.shape[0]):
        halo[i] = cluster[i]
    if len(ccenter) > 0:
        for idx in ccenter.keys():
            bord_rho[idx] = 0.0
        for i in range(1, rho.shape[0] - 1):
            for j in range(i + 1, rho.shape[0]):
                if cluster[i] != cluster[j] and distances[i - 1, j - 1] <= dc:
                    rho_aver = (rho[i] + rho[j]) / 2.0
                    if rho_aver > bord_rho[cluster[i]]:
                        bord_rho[cluster[i]] = rho_aver
                    if rho_aver > bord_rho[cluster[j]]:
                        bord_rho[cluster[j]] = rho_aver
        for i in range(1, rho.shape[0]):
            if rho[i] < bord_rho[cluster[i]]:
                halo[i] = 0
    for i in range(1, rho.shape[0]):
        if halo[i] == 0:
            cluster[i] = -1

    return cluster


def run_density_peak(
    X, density_threshold=1, distance_threshold=1, auto_select_dc=False
):
    clusters = calculate_clusters(
        X, density_threshold, distance_threshold, auto_select_dc=auto_select_dc
    )
    return list(clusters.values())



import numpy as np


def convert_to_array(lst):
    """
    Convert a list of values to a NumPy array of indices based on a sorted unique set of values.

    Args:
        lst (list): The input list of values to be converted to indices.

    Returns:
        np.ndarray: The NumPy array of indices corresponding to the input list of values.
    """
    value_to_index = {value: idx for idx, value in enumerate(sorted(set(lst)))}
    converted_list = [value_to_index[value] for value in lst]
    return np.array(converted_list)
